import java.lang.reflect.Field;
import java.lang.ref.WeakReference;

public class FinalizeTrigger {
	/**
	 * To use finalizers on PyObjects, read the documentation of FinalizablePyObject.
	 */

	public static boolean preventableMode = false;
	
	static class WeakResurrector {
		WeakReference<FinalizablePyObject> toResurrect;
		
		public WeakResurrector(FinalizablePyObject toResurrect) {
			this.toResurrect = new WeakReference<FinalizablePyObject>(toResurrect);
		}
		
		protected void finalize() throws Throwable {
			FinalizablePyObject ref = toResurrect.get();
			if (ref != null)
				FinalizeTrigger.resurrectObject(ref);
		}
	}

	static class PreventableFinalizeTrigger extends FinalizeTrigger {
		protected boolean allowFinalize = true, triggerFinalized = false;
		
		public PreventableFinalizeTrigger(FinalizablePyObject toFinalize) {
			super(toFinalize);
		}
		
		public PreventableFinalizeTrigger(FinalizablePyObject toFinalize, boolean allowFinalize) {
			super(toFinalize);
			this.allowFinalize = allowFinalize;
		}
		
		public boolean preventFinalize() {
			if (triggerFinalized && allowFinalize) return false;
			allowFinalize = false;
			return true;
		}

		public void allowFinalize() {
			if (!allowFinalize && triggerFinalized)
				((PreventableFinalizeTrigger) FinalizeTrigger.getFinalizeTrigger(toFinalize)).allowFinalize();
			allowFinalize = true;
		}
		
		protected void finalize() throws Throwable {
			if (allowFinalize)
				toFinalize.__del__();
			else
				FinalizeTrigger.setFinalizeTrigger(toFinalize, new PreventableFinalizeTrigger(toFinalize));
			triggerFinalized = true;
		}
	}

	public static FinalizeTrigger makeTrigger(FinalizablePyObject toFinalize) {
		if (preventableMode) {
			FinalizeTrigger tr = getFinalizeTrigger(toFinalize);
			PreventableFinalizeTrigger result = new PreventableFinalizeTrigger(toFinalize);
			if (tr != null && tr instanceof PreventableFinalizeTrigger) {
				result.allowFinalize = ((PreventableFinalizeTrigger) tr).allowFinalize;
			}
			return result;
		} else
			return new FinalizeTrigger(toFinalize);
	}

	/**
	 * Can be called, if an Object might be resurrected or not.
	 * Since this is more expensive than resurrectObject, one should
	 * use resurrectObject, if resurrection is sure.
	 */
	public static void resurrectOnDemand(FinalizablePyObject resurrect) {
		//This looks like a useless call, but its intention is to let the
		//GC invoke the WeakResurrector's finalizer next time.
		new WeakResurrector(resurrect);
	}

	/**
	 * Recreates the FinalizeTrigger of the given object. This makes sure that
	 * once the resurrected object is gc'ed again, its __del__ method will be
	 * called again. This emulates CPython-behaviour prior to 3.4.
	 */
	public static void resurrectObject(FinalizablePyObject resurrect) {
		setFinalizeTrigger(resurrect, makeTrigger(resurrect));
	}

	public static boolean preventFinalize(FinalizablePyObject object)
	{
		preventableMode = true;
		FinalizeTrigger tr = getFinalizeTrigger(object);
		if (tr instanceof PreventableFinalizeTrigger)
			return ((PreventableFinalizeTrigger) tr).preventFinalize();
		else {
			setFinalizeTrigger(tr.toFinalize, new PreventableFinalizeTrigger(tr.toFinalize, false));
			tr.toFinalize = null;
			return true;
		}
	}

	public static void allowFinalize(FinalizablePyObject object)
	{
		FinalizeTrigger tr = getFinalizeTrigger(object);
		if (tr instanceof PreventableFinalizeTrigger)
			((PreventableFinalizeTrigger) tr).allowFinalize();
	}

	public static void setFinalizeTrigger(FinalizablePyObject toFinalize, FinalizeTrigger trigger) {
		Field triggerField;
		try {
			triggerField = toFinalize.getClass().getField("finalizeTrigger");
		} catch (NoSuchFieldException nfe) {
			throw new IllegalArgumentException("FinalizablePyObject must have a field finalizeTrigger.");
		}
		try {
			triggerField.set(toFinalize, trigger);
		} catch (IllegalAccessException iae) {
			try {
				triggerField.setAccessible(true);
				triggerField.set(toFinalize, trigger);
			} catch (Exception e) {
				throw new IllegalArgumentException("finalizeTrigger in FinalizablePyObject must be accessible.");
			}
		}
	}

	public static FinalizeTrigger getFinalizeTrigger(FinalizablePyObject toFinalize) {
		Field triggerField;
		try {
			triggerField = toFinalize.getClass().getField("finalizeTrigger");
		} catch (NoSuchFieldException nfe) {
			throw new IllegalArgumentException("FinalizablePyObject must have a field finalizeTrigger.");
		}
		try {
			return (FinalizeTrigger) triggerField.get(toFinalize);
		} catch (IllegalAccessException iae) {
			try {
				triggerField.setAccessible(true);
				return (FinalizeTrigger) triggerField.get(toFinalize);
			} catch (Exception e) {
				throw new IllegalArgumentException("finalizeTrigger in FinalizablePyObject must be accessible.");
			}
		}
	}

	protected FinalizablePyObject toFinalize;

	protected FinalizeTrigger(FinalizablePyObject toFinalize) {
		this.toFinalize = toFinalize;
	}

	protected void finalize() throws Throwable {
		if (toFinalize != null) toFinalize.__del__();
	}
}
